# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import random
import traceback
import warnings
from typing import TYPE_CHECKING

import pytest
import xdist
from xdist.remote import Producer
from xdist.scheduler import LoadScheduling

if TYPE_CHECKING:
    from typing import Union, Sequence
    import xdist.dsession


class NoParallelAndLoadScheduling(LoadScheduling):
    def __init__(self, config, log=None):
        super(NoParallelAndLoadScheduling, self).__init__(config=config, log=log)
        self.first_send_tests = True
        self.noparallel_tests_done = False
        self.test_queue = []
        self.noparallel_node = 0

    def _send_tests(self, node, num):
        if self.first_send_tests:
            self.first_send_tests = False
            try:
                config_dict = json.loads(self.collection[0])
                assert type(config_dict) == dict and 'noparallel_tests' in config_dict
                self.pending.pop(0)
                self.config.noparallel_tests = config_dict['noparallel_tests']
                if self.config.noparallel_tests == 0:
                    self.noparallel_tests_done = True
                    random.shuffle(self.pending)
                else:
                    # shuffle tests without disrupting stop index for noparallel tests
                    # noparallel tests are not shuffled (yet), because some tests change state
                    # in a unittest.TestCase setUp/tearDown and expect this state in all subtests

                    # shuffle_noparallel = self.pending[:self.config.noparallel_tests]
                    shuffle_rest = self.pending[self.config.noparallel_tests:]
                    # random.shuffle(shuffle_noparallel)
                    random.shuffle(shuffle_rest)
                    # self.pending[:self.config.noparallel_tests] = shuffle_noparallel
                    self.pending[self.config.noparallel_tests:] = shuffle_rest

            except (json.JSONDecodeError, TypeError, AssertionError):
                warnings.warn('No configuration in first element: {first_elem}'.format(first_elem=self.collection[0]))
                traceback.print_exc()
                self.noparallel_tests_done = True

            if not self.noparallel_tests_done:
                self.noparallel_node = node
                # +1 because for remote: only run if we have an item and a next item
                count = min(1 + self.config.noparallel_tests, len(self.pending))
                super(NoParallelAndLoadScheduling, self)._send_tests(node=self.noparallel_node, num=count)
                return

        if not self.noparallel_tests_done:
            self.test_queue.append(dict(node=node, num=num))
            return

        super(NoParallelAndLoadScheduling, self)._send_tests(node=node, num=num)

    def empty_test_queue(self):
        if len(self.test_queue) > 0:
            for kwargs in self.test_queue:
                self._send_tests(**kwargs)
            self.test_queue = []

    def mark_test_complete(self, node, item_index, duration=0):
        if self.noparallel_tests_done:
            super(NoParallelAndLoadScheduling, self).mark_test_complete(
                node=node, item_index=item_index, duration=duration
            )
        elif node != self.noparallel_node:
            warn = 'mark_test_complete on {node} (item {item_index}), but noparallel tests not ready'.format(
                node=node, item_index=item_index
            )
            warnings.warn(warn, RuntimeWarning)
            super(NoParallelAndLoadScheduling, self).mark_test_complete(
                node=node, item_index=item_index, duration=duration
            )
        else:
            self.node2pending[node].remove(item_index)
            # 1 because +1 above
            if len(self.node2pending[node]) <= 1:
                # resume regular operation
                self.noparallel_tests_done = True
                self.empty_test_queue()

    def __str__(self):
        ex = super(NoParallelAndLoadScheduling, self).__str__()
        if self.first_send_tests:
            ex += ', initializing'
        else:
            if self.noparallel_tests_done:
                ex += ', parallel tests'
            else:
                ex += ', serial tests (n: {n}, node: {node}, queue: {queue})'.format(
                    n=self.config.noparallel_tests,
                    node=self.noparallel_node,
                    queue=self.node2pending[self.noparallel_node],
                )
            if self.test_queue:
                ex += ', parallel test queue: ' + str(self.test_queue)
        if self.pending:
            ex += ', {pending} pending test(s)'.format(pending=len(self.pending))
        else:
            ex += ', no pending tests'
        return ex


class NoParallelXDist:
    def __init__(self, config):
        self.log = Producer('noparallelxdist', enabled=config.option.debug)
        self.config = config

    @pytest.hookimpl
    def pytest_xdist_make_scheduler(self, config, log):
        dist = config.getoption('dist', default='load')

        if dist == 'rucio':
            return NoParallelAndLoadScheduling(config=config, log=log)
        else:
            dsession = config.pluginmanager.getplugin('dsession')  # type: xdist.dsession.DSession
            return dsession.pytest_xdist_make_scheduler(config=config, log=log)

    @pytest.hookimpl
    def pytest_collection_modifyitems(
            self,
            config,
            items: "Sequence[Union[pytest.Item, pytest.Collector]]",
    ):
        if hasattr(config, 'workerinput') or config.getoption('numprocesses', 0) > 1:
            # then xdist is running tests parallel
            config.noparallel_tests = 0

            def count_and_sort_noparallel(
                    node: "Union[pytest.Item, pytest.Collector]"
            ):
                # sort noparallel tests before any other
                noparallelmark = list(node.iter_markers('noparallel'))
                if len(noparallelmark) > 0:
                    self.log(
                        'found noparallel test:',
                        '; '.join(map(lambda mark: mark.kwargs['reason'], noparallelmark)),
                        node,
                    )
                    config.noparallel_tests += 1
                    return 0
                else:
                    return 1

            # write noparallel tests first into items sequence
            items[:] = sorted(items, key=count_and_sort_noparallel)

    @pytest.hookimpl(hookwrapper=True, tryfirst=True)
    def pytest_collection_finish(
        self,
        session  # type: pytest.Session
    ):
        if xdist.is_xdist_worker(self):
            config_str = json.dumps({'noparallel_tests': self.config.noparallel_tests})
            config_item = pytest.Item.from_parent(parent=session, name='configuration item', nodeid=config_str)
            session.items[:] = [config_item] + session.items
        yield



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import itertools
import logging
from typing import TYPE_CHECKING

from dogpile.cache.api import NoValue
from sqlalchemy import select, false

from rucio.common.utils import PriorityQueue
from rucio.common.cache import make_region_memcached
from rucio.common.config import config_get_int
from rucio.common.exception import NoDistance, RSEProtocolNotSupported, InvalidRSEExpression
from rucio.core.rse import RseCollection, list_rses
from rucio.core.config import get as core_config_get
from rucio.core.rse_expression_parser import parse_expression
from rucio.db.sqla import models
from rucio.db.sqla.session import read_session, transactional_session
from rucio.rse import rsemanager as rsemgr

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, List, Optional, Set
    from sqlalchemy.orm import Session

    LoggerFunction = Callable[..., Any]

REGION = make_region_memcached(expiration_time=600)


class Topology:
    """
    Helper private class used to dynamically load and cache the rse information
    """
    def __init__(
            self,
            rse_collection: "RseCollection",
            multihop_rses: "Set[str]",
            restricted_read_rses: "Optional[Set[str]]" = None,
            restricted_write_rses: "Optional[Set[str]]" = None,
            unavailable_read_rses: "Optional[Set[str]]" = None,
            unavailable_write_rses: "Optional[Set[str]]" = None,
    ):
        self.rse_collection = rse_collection
        self.multihop_rses = multihop_rses
        self.restricted_read_rses = restricted_read_rses or set()
        self.restricted_write_rses = restricted_write_rses or set()
        self.unavailable_read_rses = unavailable_read_rses or set()
        self.unavailable_write_rses = unavailable_write_rses or set()

    @classmethod
    @read_session
    def create_from_config(cls, ignore_availability: bool = False, *, session: "Session", logger: "LoggerFunction" = logging.log):

        include_multihop = core_config_get('transfers', 'use_multihop', default=False, expiration_time=600, session=session)

        multihop_rses = set()
        if include_multihop:
            try:
                multihop_rses = {rse['id'] for rse in parse_expression('available_for_multihop=true', session=session)}
            except InvalidRSEExpression:
                pass

        restricted_read_rses = set()
        try:
            restricted_read_rses = {rse['id'] for rse in parse_expression('restricted_read=true', session=session)}
        except InvalidRSEExpression:
            pass

        restricted_write_rses = set()
        try:
            restricted_write_rses = {rse['id'] for rse in parse_expression('restricted_write=true', session=session)}
        except InvalidRSEExpression:
            pass

        unavailable_read_rses = set()
        unavailable_write_rses = set()
        if not ignore_availability:
            unavailable_read_rses = _get_unavailable_rse_ids(operation='read', session=session, logger=logger)
            unavailable_write_rses = _get_unavailable_rse_ids(operation='write', session=session, logger=logger)
            # Disallow multihop via blocklisted RSEs
            multihop_rses = multihop_rses.difference(unavailable_write_rses).difference(unavailable_read_rses)

        topology = cls(
            rse_collection=RseCollection(),
            multihop_rses=multihop_rses,
            restricted_read_rses=restricted_read_rses,
            restricted_write_rses=restricted_write_rses,
            unavailable_read_rses=unavailable_read_rses,
            unavailable_write_rses=unavailable_write_rses,
        )
        return topology

    @transactional_session
    def search_shortest_paths(
            self,
            source_rse_ids: "List[str]",
            dest_rse_id: "str",
            operation_src: str,
            operation_dest: str,
            domain: str,
            limit_dest_schemes: "List[str]",
            inbound_links_by_node: "Optional[Dict[str, Dict[str, str]]]" = None,
            *, session: "Session"
    ) -> "Dict[str, List[Dict[str, Any]]]":
        """
        Find the shortest paths from multiple sources towards dest_rse_id.
        Does a Backwards Dijkstra's algorithm: start from destination and follow inbound links towards the sources.
        If multihop is disabled, stop after analysing direct connections to dest_rse. Otherwise, stops when all
        sources where found or the graph was traversed in integrality.

        The inbound links retrieved from the database can be accumulated into the inbound_links_by_node, passed
        from the calling context. To be able to reuse them.
        """
        HOP_PENALTY = config_get_int('transfers', 'hop_penalty', default=10, session=session)  # Penalty to be applied to each further hop

        self.rse_collection.ensure_loaded(itertools.chain(source_rse_ids, [dest_rse_id], self.multihop_rses),
                                          load_attributes=True, load_info=True, session=session)
        if self.multihop_rses:
            # Filter out island source RSEs
            sources_to_find = {rse_id for rse_id in source_rse_ids if _load_outgoing_distances_node(rse_id=rse_id, session=session)}
        else:
            sources_to_find = set(source_rse_ids)

        next_hop = {dest_rse_id: {'cumulated_distance': 0}}
        priority_q = PriorityQueue()

        remaining_sources = copy.copy(sources_to_find)
        priority_q[dest_rse_id] = 0
        while priority_q:
            current_node = priority_q.pop()

            if current_node in remaining_sources:
                remaining_sources.remove(current_node)
            if not remaining_sources:
                # We found the shortest paths to all desired sources
                break

            current_distance = next_hop[current_node]['cumulated_distance']
            inbound_links = _load_inbound_distances_node(rse_id=current_node, session=session)
            if inbound_links_by_node is not None:
                inbound_links_by_node[current_node] = inbound_links
            for adjacent_node, link_distance in sorted(inbound_links.items(),
                                                       key=lambda item: 0 if item[0] in sources_to_find else 1):
                if link_distance is None:
                    continue

                if adjacent_node not in remaining_sources and adjacent_node not in self.multihop_rses:
                    continue

                try:
                    hop_penalty = int(self.rse_collection[adjacent_node].attributes.get('hop_penalty', HOP_PENALTY))
                except ValueError:
                    hop_penalty = HOP_PENALTY
                new_adjacent_distance = current_distance + link_distance + hop_penalty
                if next_hop.get(adjacent_node, {}).get('cumulated_distance', 9999) <= new_adjacent_distance:
                    continue

                try:
                    matching_scheme = rsemgr.find_matching_scheme(
                        rse_settings_src=self.rse_collection[adjacent_node].info,
                        rse_settings_dest=self.rse_collection[current_node].info,
                        operation_src=operation_src,
                        operation_dest=operation_dest,
                        domain=domain,
                        scheme=limit_dest_schemes if adjacent_node == dest_rse_id and limit_dest_schemes else None
                    )
                    next_hop[adjacent_node] = {
                        'source_rse_id': adjacent_node,
                        'dest_rse_id': current_node,
                        'source_scheme': matching_scheme[1],
                        'dest_scheme': matching_scheme[0],
                        'source_scheme_priority': matching_scheme[3],
                        'dest_scheme_priority': matching_scheme[2],
                        'hop_distance': link_distance,
                        'cumulated_distance': new_adjacent_distance,
                    }
                    priority_q[adjacent_node] = new_adjacent_distance
                except RSEProtocolNotSupported:
                    if next_hop.get(adjacent_node) is None:
                        next_hop[adjacent_node] = {}

            if not self.multihop_rses:
                # Stop after the first iteration, which finds direct connections to destination
                break

        paths = {}
        for rse_id in source_rse_ids:
            hop = next_hop.get(rse_id)
            if hop is None:
                continue

            path = []
            while hop.get('dest_rse_id'):
                path.append(hop)
                hop = next_hop[hop['dest_rse_id']]
            paths[rse_id] = path
        return paths


@transactional_session
def get_hops(
        source_rse_id: str,
        dest_rse_id: str,
        multihop_rses: "Optional[Set[str]]" = None,
        limit_dest_schemes: "Optional[List[str]]" = None,
        *, session: "Session",
):
    """
    Get a list of hops needed to transfer date from source_rse_id to dest_rse_id.
    Ideally, the list will only include one item (dest_rse_id) since no hops are needed.
    :param source_rse_id:       Source RSE id of the transfer.
    :param dest_rse_id:         Dest RSE id of the transfer.
    :param multihop_rses:       List of RSE ids that can be used for multihop. If empty, multihop is disabled.
    :param limit_dest_schemes:  List of destination schemes the matching scheme algorithm should be limited to for a single hop.
    :returns:                   List of hops in the format [{'source_rse_id': source_rse_id, 'source_scheme': 'srm', 'source_scheme_priority': N, 'dest_rse_id': dest_rse_id, 'dest_scheme': 'srm', 'dest_scheme_priority': N}]
    :raises:                    NoDistance
    """
    if not limit_dest_schemes:
        limit_dest_schemes = []

    if not multihop_rses:
        multihop_rses = set()

    topology = Topology(rse_collection=RseCollection(), multihop_rses=multihop_rses)
    shortest_paths = topology.search_shortest_paths(source_rse_ids=[source_rse_id], dest_rse_id=dest_rse_id,
                                                    operation_src='third_party_copy_read', operation_dest='third_party_copy_write',
                                                    domain='wan', limit_dest_schemes=limit_dest_schemes, session=session)

    result = REGION.get('get_hops_dist_%s_%s_%s' % (str(source_rse_id), str(dest_rse_id), ''.join(sorted(limit_dest_schemes))))
    if not isinstance(result, NoValue):
        return result

    path = shortest_paths.get(source_rse_id)
    if path is None:
        raise NoDistance()

    if not path:
        raise RSEProtocolNotSupported()

    REGION.set('get_hops_dist_%s_%s_%s' % (str(source_rse_id), str(dest_rse_id), ''.join(sorted(limit_dest_schemes))), path)
    return path


@transactional_session
def _load_outgoing_distances_node(rse_id: str, *, session: "Session"):
    """
    Loads the outgoing edges of the distance graph for one node.
    :param rse_id:    RSE id to load the edges for.
    :param session:   The DB Session to use.
    :returns:         Dictionary based graph object.
    """

    result = REGION.get('outgoing_edges_%s' % str(rse_id))
    if isinstance(result, NoValue):
        outgoing_edges = {}
        stmt = select(
            models.Distance
        ).join(
            models.RSE,
            models.RSE.id == models.Distance.dest_rse_id
        ).where(
            models.Distance.src_rse_id == rse_id,
            models.RSE.deleted == false()
        )
        for distance in session.execute(stmt).scalars():
            if distance.distance is None:
                continue
            outgoing_edges[distance.dest_rse_id] = distance.distance if distance.distance >= 0 else 0
        REGION.set('outgoing_edges_%s' % str(rse_id), outgoing_edges)
        result = outgoing_edges
    return result


@transactional_session
def _load_inbound_distances_node(rse_id: str, *, session: "Session"):
    """
    Loads the inbound edges of the distance graph for one node.
    :param rse_id:    RSE id to load the edges for.
    :param session:   The DB Session to use.
    :returns:         Dictionary based graph object.
    """

    result = REGION.get('inbound_edges_%s' % str(rse_id))
    if isinstance(result, NoValue):
        inbound_edges = {}
        stmt = select(
            models.Distance
        ).join(
            models.RSE,
            models.RSE.id == models.Distance.src_rse_id
        ).where(
            models.Distance.dest_rse_id == rse_id,
            models.RSE.deleted == false()
        )
        for distance in session.execute(stmt).scalars():
            if distance.distance is None:
                continue
            inbound_edges[distance.src_rse_id] = distance.distance if distance.distance >= 0 else 0
        REGION.set('inbound_edges_%s' % str(rse_id), inbound_edges)
        result = inbound_edges
    return result


@read_session
def _get_unavailable_rse_ids(operation: str, *, session: "Session", logger: "LoggerFunction" = logging.log):
    """
    :param logger:   Optional decorated logger that can be passed from the calling daemons or servers.
    Get unavailable rse ids for a given operation : read, write, delete
    """

    if operation not in ['read', 'write', 'delete']:
        logger(logging.ERROR, "Wrong operation specified : %s" % operation)
        return set()
    key = 'unavailable_%s_rse_ids' % operation
    result = REGION.get(key)
    if isinstance(result, NoValue):
        try:
            logger(logging.DEBUG, "Refresh unavailable %s rses" % operation)
            availability_key = 'availability_%s' % operation
            unavailable_rses = list_rses(filters={availability_key: False}, session=session)
            unavailable_rse_ids = [rse['id'] for rse in unavailable_rses]
            REGION.set(key, unavailable_rse_ids)
            return set(unavailable_rse_ids)
        except Exception:
            logger(logging.ERROR, "Failed to refresh unavailable %s rses, error" % operation, exc_info=True)
            return set()
    return set(result)



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Dict, Any


def import_extras(module_list: "List[str]") -> "Dict[str, Any]":
    out = dict()
    for mod in module_list:
        out[mod] = None
        try:
            with warnings.catch_warnings():
                # TODO: remove when https://github.com/paramiko/paramiko/issues/2038 is fixed
                warnings.filterwarnings('ignore', 'Blowfish has been deprecated', module='paramiko')
                # TODO: deprecated python 2 and 3.6 too ...
                warnings.filterwarnings('ignore', 'Python .* is no longer supported', module='paramiko')
                out[mod] = importlib.import_module(mod)
        except ImportError:
            pass
    return out



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' new attr account table '''

import datetime

import sqlalchemy as sa

from alembic import context
from alembic.op import (create_table, create_primary_key, create_foreign_key,
                        create_check_constraint, create_index, drop_table)


# Alembic revision identifiers
revision = '4c3a4acfe006'
down_revision = '25fc855625cf'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        create_table('account_attr_map',
                     sa.Column('account', sa.String(25)),
                     sa.Column('key', sa.String(255)),
                     sa.Column('value', sa.String(255)),
                     sa.Column('created_at', sa.DateTime, default=datetime.datetime.utcnow),
                     sa.Column('updated_at', sa.DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow))

        create_primary_key('ACCOUNT_ATTR_MAP_PK', 'account_attr_map', ['account', 'key'])
        create_check_constraint('ACCOUNT_ATTR_MAP_CREATED_NN', 'account_attr_map', 'created_at is not null')
        create_check_constraint('ACCOUNT_ATTR_MAP_UPDATED_NN', 'account_attr_map', 'updated_at is not null')
        create_foreign_key('ACCOUNT_ATTR_MAP_ACCOUNT_FK', 'account_attr_map', 'accounts', ['account'], ['account'])
        create_index('ACCOUNT_ATTR_MAP_KEY_VALUE_IDX', 'account_attr_map', ['key', 'value'])


def downgrade():
    '''
    Downgrade the database to the previous revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        drop_table('account_attr_map')

    elif context.get_context().dialect.name == 'postgresql':
        # drop_constraint('ACCOUNT_ATTR_MAP_PK', 'account_attr_map', type_='primary')
        # drop_constraint('ACCOUNT_ATTR_MAP_CREATED_NN', 'account_attr_map')
        # drop_constraint('ACCOUNT_ATTR_MAP_UPDATED_NN', 'account_attr_map')
        # drop_constraint('ACCOUNT_ATTR_MAP_ACCOUNT_FK', 'account_attr_map')
        # drop_index('ACCOUNT_ATTR_MAP_KEY_VALUE_IDX', 'account_attr_map')
        # drop_table('account_attr_map')
        pass



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from rucio.api.permission import has_permission
from rucio.api.scope import list_scopes

from rucio.core.rse import get_rse_id
from rucio.core import dirac
from rucio.common.exception import AccessDenied
from rucio.common.utils import extract_scope
from rucio.db.sqla.session import transactional_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


@transactional_session
def add_files(lfns, issuer, ignore_availability, vo='def', *, session: "Session"):
    """
    Bulk add files :
    - Create the file and replica.
    - If doesn't exist create the dataset containing the file as well as a rule on the dataset on ANY sites.
    - Create all the ascendants of the dataset if they do not exist

    :param lfns: List of lfn (dictionary {'lfn': <lfn>, 'rse': <rse>, 'bytes': <bytes>, 'adler32': <adler32>, 'guid': <guid>, 'pfn': <pfn>}
    :param issuer: The issuer account.
    :param ignore_availability: A boolean to ignore blocked sites.
    :param vo: The VO to act on.
    :param session: The database session in use.

    """
    scopes = list_scopes(vo=vo, session=session)
    dids = []
    rses = {}
    for lfn in lfns:
        scope, name = extract_scope(lfn['lfn'], scopes)
        dids.append({'scope': scope, 'name': name})
        rse = lfn['rse']
        if rse not in rses:
            rse_id = get_rse_id(rse=rse, vo=vo, session=session)
            rses[rse] = rse_id
        lfn['rse_id'] = rses[rse]

    # Check if the issuer can add dids and use skip_availabitlity
    for rse in rses:
        rse_id = rses[rse]
        kwargs = {'rse': rse, 'rse_id': rse_id}
        if not has_permission(issuer=issuer, action='add_replicas', kwargs=kwargs, vo=vo, session=session):
            raise AccessDenied('Account %s can not add file replicas on %s for VO %s' % (issuer, rse, vo))
        if not has_permission(issuer=issuer, action='skip_availability_check', kwargs=kwargs, vo=vo, session=session):
            ignore_availability = False

    # Check if the issuer can add the files
    kwargs = {'issuer': issuer, 'dids': dids}
    if not has_permission(issuer=issuer, action='add_dids', kwargs=kwargs, vo=vo, session=session):
        raise AccessDenied('Account %s can not bulk add data identifier for VO %s' % (issuer, vo))

    dirac.add_files(lfns=lfns, account=issuer, ignore_availability=ignore_availability, vo=vo, session=session)



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import datetime
import hmac
import time
from hashlib import sha1

import boto3
from botocore.client import Config
from dogpile.cache.api import NO_VALUE
from google.oauth2.service_account import Credentials
from urllib.parse import urlparse, urlencode

from rucio.common.cache import make_region_memcached
from rucio.common.config import config_get, get_rse_credentials
from rucio.common.exception import UnsupportedOperation
from rucio.core.monitor import MetricManager
from rucio.core.rse import get_rse_attribute

CREDS_GCS = None

REGION = make_region_memcached(expiration_time=900)
METRICS = MetricManager(module=__name__)


def get_signed_url(rse_id: str, service: str, operation: str, url: str, lifetime=600) -> str:
    """
    Get a signed URL for a particular service and operation.

    The signed URL will be valid for 1 hour but can be overriden.

    :param rse_id: The ID of the RSE that the URL points to.
    :param service: The service to authorise, either 'gcs', 's3' or 'swift'.
    :param operation: The operation to sign, either 'read', 'write', or 'delete'.
    :param url: The URL to sign.
    :param lifetime: Lifetime of the signed URL in seconds.
    :returns: Signed URL as a variable-length string.
    """

    global CREDS_GCS

    if service not in ['gcs', 's3', 'swift']:
        raise UnsupportedOperation('Service must be "gcs", "s3" or "swift"')

    if operation not in ['read', 'write', 'delete']:
        raise UnsupportedOperation('Operation must be "read", "write", or "delete"')

    if url is None or url == '':
        raise UnsupportedOperation('URL must not be empty')

    if lifetime:
        if not isinstance(lifetime, int):
            try:
                lifetime = int(lifetime)
            except:
                raise UnsupportedOperation('Lifetime must be convertible to numeric.')

    if service == 'gcs':
        if not CREDS_GCS:
            CREDS_GCS = Credentials.from_service_account_file(config_get('credentials', 'gcs',
                                                                         raise_exception=False,
                                                                         default='/opt/rucio/etc/google-cloud-storage-test.json'))
        components = urlparse(url)
        host = components.netloc

        # select the correct operation
        operations = {'read': 'GET', 'write': 'PUT', 'delete': 'DELETE'}
        operation = operations[operation]

        # special case to test signature, force epoch time
        if lifetime is None:
            lifetime = 0
        else:
            # GCS is timezone-sensitive, don't use UTC
            # has to be converted to Unixtime
            lifetime = datetime.datetime.now() + datetime.timedelta(seconds=lifetime)
            lifetime = int(time.mktime(lifetime.timetuple()))

        # sign the path only
        path = components.path

        # assemble message to sign
        to_sign = "%s\n\n\n%s\n%s" % (operation, lifetime, path)

        # create URL-capable signature
        # first character is always a '=', remove it
        signature = urlencode({'': base64.b64encode(CREDS_GCS.sign_bytes(to_sign))})[1:]

        # assemble final signed URL
        signed_url = (
            f'https://{host}{path}'
            f'?GoogleAccessId={CREDS_GCS.service_account_email}'
            f'&Expires={lifetime}&Signature={signature}'
        )

    elif service == 's3':

        # get RSE S3 URL style (path or host)
        # path-style: https://s3.region-code.amazonaws.com/bucket-name/key-name
        # host-style: https://bucket-name.s3.region-code.amazonaws.com/key-name
        s3_url_style = get_rse_attribute(rse_id, 's3_url_style')

        # no S3 URL style specified, assume path-style
        if s3_url_style is None:
            s3_url_style = "path"

        # split URL to get hostname, bucket and key
        components = urlparse(url)
        host = components.netloc
        pathcomponents = components.path.split('/')
        if s3_url_style == "path":
            if len(pathcomponents) < 3:
                raise UnsupportedOperation('Not a valid Path-Style S3 URL')
            bucket = pathcomponents[1]
            key = '/'.join(pathcomponents[2:])
        elif s3_url_style == "host":
            hostcomponents = host.split('.')
            bucket = hostcomponents[0]
            if len(pathcomponents) < 2:
                raise UnsupportedOperation('Not a valid Host-Style S3 URL')
            key = '/'.join(pathcomponents[1:])
        else:
            raise UnsupportedOperation('Not a valid RSE S3 URL style (allowed values: path|host)')

        # remove port number from host if present
        colon = host.find(':')
        port = '443'
        if colon >= 0:
            port = host[colon + 1:]
            host = host[:colon]

        # look up in RSE account configuration by RSE ID
        cred_name = rse_id
        cred = REGION.get('s3-%s' % cred_name)
        if cred is NO_VALUE:
            rse_cred = get_rse_credentials()
            cred = rse_cred.get(cred_name)
            REGION.set('s3-%s' % cred_name, cred)
        access_key = cred['access_key']
        secret_key = cred['secret_key']
        signature_version = cred['signature_version']
        region_name = cred['region']

        if operation == 'read':
            s3op = 'get_object'
        elif operation == 'write':
            s3op = 'put_object'
        else:
            s3op = 'delete_object'

        with METRICS.timer('signs3'):

            if s3_url_style == "host":
                s3_url_style = "virtual"

            s3 = boto3.client(service_name='s3',
                              endpoint_url=f'https://{host}:{port}',
                              aws_access_key_id=access_key,
                              aws_secret_access_key=secret_key,
                              config=Config(signature_version=signature_version,
                                            region_name=region_name,
                                            s3={"addressing_style": s3_url_style}))

            signed_url: str = s3.generate_presigned_url(
                s3op, Params={'Bucket': bucket, 'Key': key}, ExpiresIn=lifetime)

    else:  # service == 'swift'
        # split URL to get hostname and path
        components = urlparse(url)
        host = components.netloc

        # remove port number from host if present
        colon = host.find(':')
        if colon >= 0:
            host = host[:colon]

        # use RSE ID to look up key
        cred_name = rse_id

        # look up tempurl signing key
        cred = REGION.get('swift-%s' % cred_name)
        if cred is NO_VALUE:
            rse_cred = get_rse_credentials()
            cred = rse_cred.get(cred_name)
            REGION.set('swift-%s' % cred_name, cred)
        tempurl_key = cred['tempurl_key']

        if operation == 'read':
            swiftop = 'GET'
        elif operation == 'write':
            swiftop = 'PUT'
        else:
            swiftop = 'DELETE'

        expires = int(time.time() + lifetime)

        # create signed URL
        with METRICS.timer('signswift'):
            hmac_body = u'%s\n%s\n%s' % (swiftop, expires, components.path)
            # Python 3 hmac only accepts bytes or bytearray
            sig = hmac.new(bytearray(tempurl_key, 'utf-8'), bytearray(hmac_body, 'utf-8'), sha1).hexdigest()
            signed_url = f'https://{host}{components.path}?temp_url_sig={sig}&temp_url_expires={expires}'

    return signed_url
